/*
  ESP32 CYD Calculator
  
  RPN Scientific-ish Calculator for the CYD

  There is probably a better way to code this
  but here is what I came up with.

  Hardware:
    ESP32 Cheap Yellow Display (CYD) module (ESP32-2432S028)
    
  See Hardware.h for GPIO connections

  Compile using the ESP32 Dev Module
  
  Concept, Design and Implementation: Craig A. Lindley
  Last Update: 01/10/2025
*/

#include <SPI.h>
#include <XPT2046_Touchscreen.h>
#include <Adafruit_GFX.h>
#include <Adafruit_ILI9341.h>
#include <Fonts/FreeMono9pt7b.h>
#include <stack>

#include "Hardware.h"
#include "ButtonTokens.h"
#include "TextHelper.h"
#include "DisplayHelper.h"
#include "Button.h"
#include "ButtonManager.h"

// Program version numbers
#define MAJOR_VERSION 1
#define MINOR_VERSION 0

// Misc screen attributes
#define SCREEN_ROTATION 3
#define SCREEN_COLOR ILI9341_LIGHTGREY
#define SCREEN_TEXT_COLOR ILI9341_WHITE

/****************************************************************/
/***                 Driver Object Instantiation              ***/
/****************************************************************/

// Instantiate the LCD driver
Adafruit_ILI9341 lcd(LCD_CS, LCD_DC, LCD_MOSI, LCD_SCLK, LCD_RST, LCD_MISO);

// Instantiate TextHelper
TextHelper th(&lcd);

// Instantiate DisplayHelper
DisplayHelper dh(&lcd);

// Configure VSPI interface for touch controller
SPIClass v_SPI(VSPI);

// Instantiate the touch screen driver
XPT2046_Touchscreen touchscreen(TOUCH_CS, TOUCH_IRQ);

// Instantiate the button manager
ButtonManager bm;

// The RPN stack
// Functions include
// top(), empty(), size(), push(), pop()
// NOTE: pop does not work like one would expect
std::stack<double> stack;

/****************************************************************/
/***                       Misc Variables                     ***/
/****************************************************************/

// Input string buffer
// Used in assembling a numeric value before it is committed to the stack
char inputBuffer[30];

// Variables used in inputing a number
boolean fixing;
boolean hasDP;
boolean inputing;
boolean exponentiating;
int inputBufferIndex;
int charsEntered;
char* end;

/****************************************************************/
/***                 Number Processing Functions              ***/
/****************************************************************/

// Display the number on the top of the stack
// The stack is not altered
void displayDoubleOnTOS() {
  dh.displayNumber(stack.top());
}

// Push a double number onto the stack and
// display it
void pushDoubleOntoStack(double num) {
  stack.push(num);
  displayDoubleOnTOS();
}

double popDoubleFromStack() {
  double tmp = stack.top();
  stack.pop();
  return tmp;
}

void clearInputBuffer() {
  memset(inputBuffer, 0, sizeof(inputBuffer));
  inputBufferIndex = 0;
  charsEntered = 0;
}

void resetParsingParameters() {
  inputing = true;
  fixing = false;
  hasDP = false;
  exponentiating = false;
  clearInputBuffer();
}

void addCharToInputBuffer(char ch) {
  inputBuffer[inputBufferIndex++] = ch;
  charsEntered++;
  dh.displayInputBuffer(inputBuffer);
}

void backspaceInputBuffer() {
  if (charsEntered > 0) {
    charsEntered--;
    inputBuffer[--inputBufferIndex] = 0;
    dh.displayInputBuffer(inputBuffer);
  }
}

// Simulate pressing the ENTER key
void doEnter() {
  if (inputBufferIndex != 0) {
    pushDoubleOntoStack(strtod(inputBuffer, &end));
    resetParsingParameters();
  }
}

// Process a number key press
void processNumberKeyPress(int dPlaces, char asciiCode) {
  if (fixing) {
    dh.setDecimalPlaces(dPlaces);
    displayDoubleOnTOS();
    fixing = false;
  } else {
    if (inputing) {
      clearInputBuffer();
      dh.clearDisplay();
      inputing = false;
    }
    addCharToInputBuffer(asciiCode);
  }
}

/****************************************************************/
/***                       Misc Functions                     ***/
/****************************************************************/

// Calculate y pixel position corresponding to a text line
// for the specified font and text size. The selected font
// allows for 12 lines on the landscape screen 0 .. 11
int calcLineOffset(int line) {
  return (((line + 1) * 1.5) * th.getTextHeight("gMy"));
}

// Display welcome screen
void displayWelcomeScreen(void) {

  char buffer[20];
  sprintf(buffer, "Version: %d.%d", MAJOR_VERSION, MINOR_VERSION);

  // Clear screen
  lcd.fillScreen(ILI9341_BLACK);
  lcd.setTextColor(ILI9341_LIGHTGREY);

  // Display credits
  th.drawCenteredText(calcLineOffset(0), "Cheap Yellow Display");
  th.drawCenteredText(calcLineOffset(1), "RPN");
  th.drawCenteredText(calcLineOffset(2), "Scientific-ish Calculator");
  th.drawCenteredText(calcLineOffset(3), buffer);
  th.drawCenteredText(calcLineOffset(7), "Written By");
  th.drawCenteredText(calcLineOffset(8), "Craig A. Lindley");
  th.drawCenteredText(calcLineOffset(9), "January 2025");

  delay(2000);

  // Clear screen
  lcd.fillScreen(SCREEN_COLOR);
  lcd.setTextColor(SCREEN_TEXT_COLOR);
}

// Run the calculator state machine
void runCalculatorSM(enum BTOKEN token) {

  double tmp1, tmp2;

  int stackDepth = stack.size();


  switch (token) {

    case T_SR:
      {
        doEnter();  // Simulate pressing the ENTER key
        if (stackDepth > 0) {
          tmp1 = popDoubleFromStack();
          pushDoubleOntoStack(sqrt(tmp1));
        }
      }
      break;

    case T_10X:
      {
        doEnter();  // Simulate pressing the ENTER key
        if (stackDepth > 0) {
          tmp1 = popDoubleFromStack();
          pushDoubleOntoStack(pow(10, tmp1));
        }
      }
      break;

    case T_YX:
      {
        doEnter();  // Simulate pressing the ENTER key
        if (stackDepth > 1) {
          tmp1 = popDoubleFromStack();
          tmp2 = popDoubleFromStack();
          pushDoubleOntoStack(pow(tmp2, tmp1));
        }
      }
      break;

    case T_CHS:
      {
        if (exponentiating) {
          addCharToInputBuffer('-');
        } else {
          doEnter();  // Simulate pressing the ENTER key
          if (stackDepth > 0) {
            tmp1 = popDoubleFromStack();
            pushDoubleOntoStack(tmp1 * -1.0);
          }
        }
      }
      break;

    case T_7:
      {
        processNumberKeyPress(7, '7');
      }
      break;

    case T_8:
      {
        processNumberKeyPress(8, '8');
      }
      break;

    case T_9:
      {
        processNumberKeyPress(9, '9');
      }
      break;

    case T_DIV:
      {
        doEnter();  // Simulate pressing the ENTER key
        if (stackDepth > 1) {
          tmp1 = popDoubleFromStack();
          tmp2 = popDoubleFromStack();
          pushDoubleOntoStack(tmp2 / tmp1);
        }
      }
      break;

    case T_SIN:
      {
        doEnter();  // Simulate pressing the ENTER key
        if (stackDepth > 0) {
          tmp1 = popDoubleFromStack();
          pushDoubleOntoStack(sin(tmp1));
        }
      }
      break;

    case T_COS:
      {
        doEnter();  // Simulate pressing the ENTER key
        if (stackDepth > 0) {
          tmp1 = popDoubleFromStack();
          pushDoubleOntoStack(cos(tmp1));
        }
      }
      break;

    case T_TAN:
      {
        doEnter();  // Simulate pressing the ENTER key
        if (stackDepth > 0) {
          tmp1 = popDoubleFromStack();
          pushDoubleOntoStack(tan(tmp1));
        }
      }
      break;

    case T_EEX:
      {
        if (!exponentiating) {
          addCharToInputBuffer('e');
          exponentiating = true;
        }
      }
      break;

    case T_4:
      {
        processNumberKeyPress(4, '4');
      }
      break;

    case T_5:
      {
        processNumberKeyPress(5, '5');
      }
      break;

    case T_6:
      {
        processNumberKeyPress(6, '6');
      }
      break;

    case T_MULT:
      {
        doEnter();  // Simulate pressing the ENTER key
        if (stackDepth > 1) {
          tmp1 = popDoubleFromStack();
          tmp2 = popDoubleFromStack();
          pushDoubleOntoStack(tmp1 * tmp2);
        }
      }
      break;

    case T_DTOR:
      {
        doEnter();  // Simulate pressing the ENTER key
        if (stackDepth > 0) {
          tmp1 = popDoubleFromStack();
          pushDoubleOntoStack(tmp1 * 0.01745329252);
        }
      }
      break;

    case T_RTOD:
      {
        doEnter();  // Simulate pressing the ENTER key
        if (stackDepth > 0) {
          tmp1 = popDoubleFromStack();
          pushDoubleOntoStack(tmp1 * 57.295779513);
        }
      }
      break;

    case T_FIX:
      {
        fixing = true;
      }
      break;

    case T_SWP:
      {
        doEnter();  // Simulate pressing the ENTER key
        if (stackDepth > 1) {
          tmp1 = popDoubleFromStack();
          tmp2 = popDoubleFromStack();
          stack.push(tmp1);
          pushDoubleOntoStack(tmp2);
        }
      }
      break;

    case T_1:
      {
        processNumberKeyPress(1, '1');
      }
      break;

    case T_2:
      {
        processNumberKeyPress(2, '2');
      }
      break;

    case T_3:
      {
        processNumberKeyPress(3, '3');
      }
      break;

    case T_MINUS:
      {
        doEnter();  // Simulate pressing the ENTER key
        if (stackDepth > 1) {
          tmp1 = popDoubleFromStack();
          tmp2 = popDoubleFromStack();
          pushDoubleOntoStack(tmp2 - tmp1);
        }
      }
      break;

    case T_MMTOI:
      {
        doEnter();  // Simulate pressing the ENTER key
        if (stackDepth > 0) {
          tmp1 = popDoubleFromStack();
          pushDoubleOntoStack(tmp1 / 25.4);
        }
      }
      break;

    case T_ITOMM:
      {
        doEnter();  // Simulate pressing the ENTER key
        if (stackDepth > 0) {
          tmp1 = popDoubleFromStack();
          pushDoubleOntoStack(tmp1 * 25.4);
        }
      }
      break;

    case T_CLR:
      {
        while (stack.size() != 0) {
          stack.pop();
        }
        // Push a zero onto the empty stack
        pushDoubleOntoStack(0.0);
      }
      break;

    case T_ENT:
      {
        doEnter();
      }
      break;

    case T_0:
      {
        // Cannot fix 0 decimal places so install default of 2
        processNumberKeyPress(2, '0');
      }
      break;

    case T_DOT:
      {
        if (!hasDP) {
          if (inputing) {
            dh.clearDisplay();
            inputing = false;
          }
          addCharToInputBuffer('.');
          hasDP = true;
        }
      }
      break;

    case T_BS:
      {
        backspaceInputBuffer();
      }
      break;

    case T_PLUS:
      {
        Serial.printf("SS: %d\n", stack.size());

        doEnter();  // Simulate pressing the ENTER key
        if (stackDepth > 1) {
          tmp1 = popDoubleFromStack();
          tmp2 = popDoubleFromStack();
          pushDoubleOntoStack(tmp2 + tmp1);
        }
        Serial.printf("SS: %d\n", stack.size());
      }
      break;
  }
  // debug(token);
}

/****************************************************************/
/***                        Program Setup                     ***/
/****************************************************************/

void setup() {

  Serial.begin(115200);
  Serial.println("\n\nStarting Up");

  // Initialize LCD ILI9341 display controller
  lcd.begin();
  lcd.setRotation(SCREEN_ROTATION);
  lcd.setFont(&FreeMono9pt7b);
  lcd.setTextSize(1);
  lcd.fillScreen(SCREEN_COLOR);
  lcd.setTextColor(SCREEN_TEXT_COLOR);

  // Configure and turn on LCD backlight
  pinMode(LCD_BL, OUTPUT);
  digitalWrite(LCD_BL, HIGH);

  // Initialize touch screen controller
  v_SPI.begin(TOUCH_CLK, TOUCH_MISO, TOUCH_MOSI, TOUCH_CS);
  touchscreen.begin(v_SPI);
  touchscreen.setRotation(3);

  // Display the welcome screen
  displayWelcomeScreen();

  // Draw the calc display
  dh.clearDisplay();

  // Draw the calc buttons
  bm.drawButtons(&lcd);

  // Push a zero onto the stack
  pushDoubleOntoStack(0.0);

  // Resest numeric parsing variables
  resetParsingParameters();
}

/****************************************************************/
/***                        Program Loop                      ***/
/****************************************************************/

POINT pt;

void loop() {

  // Checks if screen touched
  if (touchscreen.tirqTouched() && touchscreen.touched()) {
    // Yes it was, get raw location
    TS_Point p = touchscreen.getPoint();
    // Calibrate Touchscreen points with map function to the correct width and height
    pt.x = map(p.x, 200, 3700, 1, 320);
    pt.y = map(p.y, 240, 3800, 1, 240);

    enum BTOKEN token = bm.pollAllButtons(&lcd, pt);
    if (token != T_NONE) {
      runCalculatorSM(token);
    }
  }

  delay(100);
}
